import sys
import time

from je_auto_control.utils.exception.exceptions import AutoControlException
from je_auto_control.utils.exception.exception_tag import linux_import_error

if sys.platform not in ["linux", "linux2"]:
    raise AutoControlException(linux_import_error)

from je_auto_control.linux_with_x11.core.utils.x11_linux_display import display
from Xlib.ext.xtest import fake_input
from Xlib import X


def press_key(keycode: int):
    """
    :param keycode which keycode we want to press
    """
    try:
        time.sleep(0.01)
        fake_input(display, X.KeyPress, keycode)
        display.sync()
    except struct.error as error:
        print(repr(error), file=sys.stderr)


def release_key(keycode: int):
    """
    :param keycode which keycode we want to release
    """
    try:
        time.sleep(0.01)
        fake_input(display, X.KeyRelease, keycode)
        display.sync()
    except struct.error as error:
        print(repr(error), file=sys.stderr)



import sys

from je_auto_control.utils.exception.exceptions import AutoControlException
from je_auto_control.utils.exception.exception_tag import linux_import_error
from je_auto_control.utils.exception.exception_tag import listener_error


if sys.platform not in ["linux", "linux2"]:
    raise AutoControlException(linux_import_error)

import Xlib.threaded

from Xlib.display import Display
from Xlib import X
from Xlib.ext import record
from Xlib.protocol import rq

from threading import Thread

# get current display
current_display = Display()


class KeypressHandler(Thread):

    def __init__(self, default_daemon: bool = True):
        """
        setDaemon : default damon is true
        still listener : continue listener keycode ?
        event_key_code : now current key code default is 0
        """
        super().__init__()
        self.setDaemon(default_daemon)
        self.still_listener = True
        self.record_flag = False
        self.record_queue = None
        self.event_keycode = 0
        self.event_position = 0, 0

    # two times because press and release
    def check_is_press(self, keycode: int):
        """
        :param keycode we want to check
        """
        if keycode == self.event_keycode:
            self.event_keycode = 0
            return True
        else:
            return False

    def run(self, reply):
        """
        :param reply listener return data
        get data
        while data not null and still listener
            get event

        """
        try:
            data = reply.data
            while len(data) and self.still_listener:
                event, data = rq.EventField(None).parse_binary_value(data, current_display.display, None, None)
                if event.detail != 0:
                    if event.type is X.ButtonRelease or event.type is X.KeyRelease:
                        self.event_keycode = event.detail
                        self.event_position = event.root_x, event.root_y
                        if self.record_flag is True:
                            temp = (event.type, event.detail, event.root_x, event.root_y)
                            self.record_queue.put(temp)
        except AutoControlException:
            raise AutoControlException(listener_error)

    def record(self, record_queue):
        """
        :param record_queue the queue test_record action
        """
        self.record_flag = True
        self.record_queue = record_queue

    def stop_record(self):
        self.record_flag = False
        return self.record_queue


class XWindowsKeypressListener(Thread):

    def __init__(self, default_daemon=True):
        """
        :param default_daemon default kill when program down
        create handler
        set root
        """
        super().__init__()
        self.setDaemon(default_daemon)
        self.still_listener = True
        self.handler = KeypressHandler()
        self.root = current_display.screen().root
        self.context = None

    def check_is_press(self, keycode: int):
        """
        :param keycode check this keycode is press?
        """
        return self.handler.check_is_press(keycode)

    def run(self):
        """
        while still listener
            get context
            set handler
            set test_record
            get event
        """
        if self.still_listener:
            try:
                # Monitor keypress and button press
                if self.context is None:
                    self.context = current_display.record_create_context(
                        0,
                        [record.AllClients],
                        [{
                            'core_requests': (0, 0),
                            'core_replies': (0, 0),
                            'ext_requests': (0, 0, 0, 0),
                            'ext_replies': (0, 0, 0, 0),
                            'delivered_events': (0, 0),
                            'device_events': (X.KeyReleaseMask, X.ButtonReleaseMask),
                            'errors': (0, 0),
                            'client_started': False,
                            'client_died': False,
                        }])
                    current_display.record_enable_context(self.context, self.handler.run)
                    current_display.record_free_context(self.context)
                # keep running this to get event
                next_event = self.root.display.next_event()
            except AutoControlException:
                raise AutoControlException(listener_error)
            finally:
                self.handler.still_listener = False
                self.still_listener = False

    def record(self, record_queue):
        self.handler.record(record_queue)

    def stop_record(self):
        return self.handler.stop_record()


xwindows_listener = XWindowsKeypressListener()
xwindows_listener.start()


def check_key_is_press(keycode: int):
    """
    :param keycode check this keycode is press?
    """
    return xwindows_listener.check_is_press(keycode)


def x11_linux_record(record_queue):
    """
    :param record_queue the queue test_record action
    """
    xwindows_listener.record(record_queue)


def x11_linux_stop_record():
    """
    stop test_record action
    """
    return xwindows_listener.stop_record()


if __name__ == "__main__":
    from queue import Queue
    test_queue = Queue()
    xwindows_listener.record(test_queue)
    while True:
        pass



from pathlib import Path


def create_dir(dir_name: str):
    Path(dir_name).mkdir(
        parents=True,
        exist_ok=True
    )


def create_template_dir():
    create_dir("auto_control/template")



import sys

from je_auto_control.utils.image import template_detection
from je_auto_control.utils.exception.exception_tag import cant_find_image
from je_auto_control.utils.exception.exception_tag import find_image_error_variable
from je_auto_control.utils.exception.exceptions import ImageNotFoundException
from je_auto_control.wrapper.auto_control_mouse import click_mouse
from je_auto_control.wrapper.auto_control_mouse import set_position
from je_auto_control.utils.image.screenshot import pil_screenshot
from je_auto_control.utils.test_record.record_test_class import record_total


def locate_all_image(image, detect_threshold: [float, int] = 1, draw_image: bool = False):
    """
     use to locate all image that detected and then return detected images list
    :param image which image we want to find on screen (png or PIL ImageGrab.grab())
    :param detect_threshold detect precision 0.0 ~ 1.0; 1 is absolute equal (float or int)
    :param draw_image draw detect tag on return image (bool)
    """
    param = locals()
    try:
        try:
            image_data_array = template_detection.find_image_multi(image, detect_threshold, draw_image)
        except ImageNotFoundException as error:
            raise ImageNotFoundException(find_image_error_variable + " " + repr(error))
        if image_data_array[0] is True:
            record_total("locate_all_image", param)
            return image_data_array[1]
        else:
            raise ImageNotFoundException(cant_find_image)
    except Exception as error:
        record_total("locate_all_image", param, repr(error))
        print(repr(error), file=sys.stderr)


def locate_image_center(image, detect_threshold: [float, int] = 1, draw_image: bool = False):
    """
    use to locate image and return image center position
    :param image which image we want to find on screen (png or PIL ImageGrab.grab())
    :param detect_threshold detect precision 0.0 ~ 1.0; 1 is absolute equal (float or int)
    :param draw_image draw detect tag on return image (bool)
    """
    param = locals()
    try:
        try:
            image_data_array = template_detection.find_image(image, detect_threshold, draw_image)
        except ImageNotFoundException as error:
            raise ImageNotFoundException(find_image_error_variable + " " + repr(error))
        if image_data_array[0] is True:
            height = image_data_array[1][2] - image_data_array[1][0]
            width = image_data_array[1][3] - image_data_array[1][1]
            center = [int(height / 2), int(width / 2)]
            record_total("locate_image_center", param)
            return [image_data_array[1][0] + center[0], image_data_array[1][1] + center[1]]
        else:
            raise ImageNotFoundException(cant_find_image)
    except Exception as error:
        record_total("locate_image_center", param, repr(error))
        print(repr(error), file=sys.stderr)


def locate_and_click(image, mouse_keycode: [int, str], detect_threshold: [float, int] = 1, draw_image: bool = False):
    """
    use to locate image and click image center position and the return image center position
    :param image which image we want to find on screen (png or PIL ImageGrab.grab())
    :param mouse_keycode which mouse keycode we want to click
    :param detect_threshold detect precision 0.0 ~ 1.0; 1 is absolute equal (float or int)
    :param draw_image draw detect tag on return image (bool)
    """
    param = locals()
    try:
        try:
            image_data_array = template_detection.find_image(image, detect_threshold, draw_image)
        except ImageNotFoundException:
            raise ImageNotFoundException(find_image_error_variable)
        if image_data_array[0] is True:
            height = image_data_array[1][2] - image_data_array[1][0]
            width = image_data_array[1][3] - image_data_array[1][1]
            center = [int(height / 2), int(width / 2)]
            image_center_x = image_data_array[1][0] + center[0]
            image_center_y = image_data_array[1][1] + center[1]
            set_position(int(image_center_x), int(image_center_y))
            click_mouse(mouse_keycode)
            record_total("locate_and_click", param)
            return [image_center_x, image_center_y]
        else:
            raise ImageNotFoundException(cant_find_image)
    except Exception as error:
        record_total("locate_and_click", param, repr(error))
        print(repr(error), file=sys.stderr)


def screenshot(file_path: str = None, region: list = None):
    """
    use to get now screen image return image
    :param file_path save screenshot path (None is no save)
    :param region screenshot region (screenshot region on screen)
    """
    param = locals()
    try:
        record_total("screenshot", param)
        return pil_screenshot(file_path, region)
    except Exception as error:
        print(repr(error), file=sys.stderr)
        record_total("screenshot", param, repr(error))



# argparse
import argparse

from je_auto_control.utils.json.json_file import read_action_json
from je_auto_control.utils.executor.action_executor import execute_action

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("-e", "--execute_file", type=str, help="choose action file to execute")
    args = parser.parse_args()
    if args.execute_file is not None:
        execute_action(read_action_json(args.execute_file))
    else:
        print("No argument, Hello There :)")



import sys

from je_auto_control.utils.exception.exception_tag import windows_import_error
from je_auto_control.utils.exception.exceptions import AutoControlException

if sys.platform not in ["win32", "cygwin", "msys"]:
    raise AutoControlException(windows_import_error)

from je_auto_control.windows.core.utils.win32_ctype_input import Input
from je_auto_control.windows.core.utils.win32_ctype_input import Keyboard
from je_auto_control.windows.core.utils.win32_ctype_input import KeyboardInput
from je_auto_control.windows.core.utils.win32_ctype_input import SendInput
from je_auto_control.windows.core.utils.win32_ctype_input import ctypes
from je_auto_control.windows.core.utils.win32_ctype_input import win32_EventF_KEYUP


def press_key(keycode: int):
    """
    :param keycode which keycode we want to press
    """
    x = Input(type=Keyboard, ki=KeyboardInput(wVk=keycode))
    SendInput(1, ctypes.byref(x), ctypes.sizeof(x))


def release_key(keycode: int):
    """
    :param keycode which keycode we want to release
    """
    x = Input(type=Keyboard, ki=KeyboardInput(wVk=keycode, dwFlags=win32_EventF_KEYUP))
    SendInput(1, ctypes.byref(x), ctypes.sizeof(x))



import sys

import cv2
import numpy as np

from je_auto_control.utils.image.screenshot import pil_screenshot
from je_auto_control.utils.exception.exception_tag import screen_get_size
from je_auto_control.utils.exception.exception_tag import screen_screenshot
from je_auto_control.utils.exception.exceptions import AutoControlScreenException
from je_auto_control.wrapper.platform_wrapper import screen
from je_auto_control.utils.test_record.record_test_class import record_total


def size():
    """
    get screen size
    """
    try:
        try:
            record_total("size", None)
            return screen.size()
        except AutoControlScreenException:
            raise AutoControlScreenException(screen_get_size)
    except Exception as error:
        record_total("size", None, repr(error))
        print(repr(error), file=sys.stderr)


def screenshot(file_path: str = None, region: list = None):
    """
    use to capture current screen image
    :param file_path screenshot file save path
    :param region screenshot region
    """
    param = locals()
    try:
        try:
            record_total("screenshot", param)
            return cv2.cvtColor(np.array(pil_screenshot(file_path=file_path, region=region)), cv2.COLOR_RGB2BGR)
        except AutoControlScreenException as error:
            raise AutoControlScreenException(screen_screenshot + " " + repr(error))
    except Exception as error:
        record_total("screenshot", None, repr(error))
        print(repr(error), file=sys.stderr)



import sys

from je_auto_control.utils.exception.exception_tag import osx_import_error
from je_auto_control.utils.exception.exceptions import AutoControlException

if sys.platform not in ["darwin"]:
    raise AutoControlException(osx_import_error)

import Quartz


def size():
    """
    get screen size
    """
    return Quartz.CGDisplayPixelsWide((Quartz.CGMainDisplayID())), Quartz.CGDisplayPixelsHigh(Quartz.CGMainDisplayID())



import datetime


class TestRecord(object):

    def __init__(self, init_total_record: bool = False):
        self.init_total_record = init_total_record
        self.total_record_list = list()

    def clean_record(self):
        self.total_record_list = list()


test_record = TestRecord()


def record_total(function_name: str, local_param, program_exception: str = None):
    if not test_record.init_total_record:
        pass
    else:
        test_record.total_record_list.append(
            {
                "function_name": function_name,
                "local_param": local_param,
                "time": str(datetime.datetime.now()),
                "program_exception": repr(program_exception)
            }
        )



import sys

from je_auto_control.utils.executor.action_executor import execute_action
from je_auto_control.utils.exception.exception_tag import macos_record_error
from je_auto_control.utils.exception.exceptions import AutoControlException
from je_auto_control.utils.exception.exceptions import AutoControlJsonActionException
from je_auto_control.wrapper.platform_wrapper import recorder
from je_auto_control.utils.test_record.record_test_class import record_total


def record():
    """
    start record keyboard and mouse event until stop_record
    """
    try:
        if sys.platform == "darwin":
            raise AutoControlException(macos_record_error)
        record_total("record", None)
        return recorder.record()
    except Exception as error:
        record_total("record", None, repr(error))
        print(repr(error), file=sys.stderr)


def stop_record():
    """
    stop current record
    """
    try:
        if sys.platform == "darwin":
            raise AutoControlException(macos_record_error)
        action_queue = recorder.stop_record()
        if action_queue is None:
            raise AutoControlJsonActionException
        action_list = list(action_queue.queue)
        new_list = list()
        for action in action_list:
            if action[0] == "type_key":
                new_list.append([action[0], dict([["keycode", action[1]]])])
            else:
                new_list.append([action[0], dict(zip(["mouse_keycode", "x", "y"], [action[0], action[1], action[2]]))])
        record_total("stop_record", None)
        return new_list
    except Exception as error:
        record_total("stop_record", None, repr(error))
        print(repr(error), file=sys.stderr)




import sys

from je_auto_control.utils.exception.exception_tag import windows_import_error
from je_auto_control.utils.exception.exceptions import AutoControlException

if sys.platform not in ["win32", "cygwin", "msys"]:
    raise AutoControlException(windows_import_error)

import ctypes

user32 = ctypes.windll.user32
user32.SetProcessDPIAware()


def size():
    """
    get screen size
    """
    return [user32.GetSystemMetrics(0), user32.GetSystemMetrics(1)]



import sys
import time

from je_auto_control.utils.exception.exceptions import AutoControlException
from je_auto_control.utils.exception.exception_tag import linux_import_error

if sys.platform not in ["linux", "linux2"]:
    raise AutoControlException(linux_import_error)

from Xlib import X
from Xlib.ext.xtest import fake_input

from je_auto_control.linux_with_x11.core.utils.x11_linux_display import display

x11_linux_mouse_left = 1
x11_linux_mouse_middle = 2
x11_linux_mouse_right = 3
x11_linux_scroll_direction_up = 4
x11_linux_scroll_direction_down = 5
x11_linux_scroll_direction_left = 6
x11_linux_scroll_direction_right = 7


def position():
    """
    get mouse current position
    """
    coord = display.screen().root.query_pointer()._data
    return coord["root_x"], coord["root_y"]


def set_position(x: int, y: int):
    """
    :param x we want to set mouse x position
    :param y we want to set mouse y position
    """
    time.sleep(0.01)
    fake_input(display, X.MotionNotify, x=x, y=y)
    display.sync()


def press_mouse(mouse_keycode: int):
    """
    :param mouse_keycode mouse keycode we want to press
    """
    time.sleep(0.01)
    fake_input(display, X.ButtonPress, mouse_keycode)
    display.sync()


def release_mouse(mouse_keycode: int):
    """
    :param mouse_keycode which mouse keycode we want to release
    """
    time.sleep(0.01)
    fake_input(display, X.ButtonRelease, mouse_keycode)
    display.sync()


def click_mouse(mouse_keycode: int, x=None, y=None):
    """
    :param mouse_keycode which mouse keycode we want to click
    :param x set mouse x position
    :param y set mouse y position
    """
    if x and y is not None:
        set_position(x, y)
    press_mouse(mouse_keycode)
    release_mouse(mouse_keycode)


def scroll(scroll_value: int, scroll_direction: int):
    """"
    :param scroll_value scroll unit
    :param scroll_direction what direction you want to scroll
    scroll_direction = 4 : direction up
    scroll_direction = 5 : direction down
    scroll_direction = 6 : direction left
    scroll_direction = 7 : direction right
    """
    for i in range(scroll_value):
        click_mouse(scroll_direction)

